//
// Created by Paul Ross on 03/11/2020.
//

#ifndef CPYMEMTRACE_PYMEMTRACE_UTIL_H
#define CPYMEMTRACE_PYMEMTRACE_UTIL_H

char *create_filename(void);
char *current_working_directory(void);

#endif //CPYMEMTRACE_PYMEMTRACE_UTIL_H
